/*
 *	input_caf.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spWave_Caf.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#define SP_CAF_INPUT_PLUGIN

#include "plugin_caf.c"

static spBool spIsSupportedByPluginCaf(const char *filename);

static spPluginError spOpenPluginCaf(void *instance, const char *filename, const char *mode);
static spBool spClosePluginCaf(void *instance);
static spPluginState spGetPluginStateCaf(void *instance);
static spLong spGetPluginCurrentPositionCaf(void *instance);
static spBool spStopPluginCaf(void *instance);
static spBool spPausePluginCaf(void *instance);
static spBool spRestartPluginCaf(void *instance);

static long spReadPluginCaf(void *instance, char *data, long length);
static spBool spSeekPluginCaf(void *instance, spLong pos);
static spLong spGetPluginTotalLengthCaf(void *instance);

static spInputPluginRec sp_input_plugin_caf = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "CAF",
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginCaf,
    spFreePluginCaf,
    "CAF Input Plugin",
    "CAF Input Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitPluginInstanceCaf,
    spFreePluginInstanceCaf,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_caf_file_type_list,
    sp_caf_file_desc_list,
    sp_caf_file_filter_list,
    spIsSupportedByPluginCaf,
    spSetPluginFileTypeCaf,  
    spGetPluginFileTypeCaf,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK
     | SP_SONG_ALBUM_ARTIST_MASK | SP_SONG_COMPOSER_MASK | SP_SONG_TRACK_TOTAL_MASK
     | SP_SONG_DISC_MASK | SP_SONG_DISC_TOTAL_MASK),
    spSetPluginSongInfoCaf,
    spGetPluginSongInfoCaf,
    spGetPluginBestSuffixCaf,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
    
    spSetPluginSampleBitCaf,   
    spGetPluginSampleBitCaf,   
    spSetPluginChannelCaf,     
    spGetPluginChannelCaf,     
    spSetPluginSampleRateCaf,  
    spGetPluginSampleRateCaf,  
    spSetPluginOtherInfoCaf,   
    spGetPluginOtherInfoCaf,
    
    spOpenPluginCaf,              
    spClosePluginCaf,             
    spGetPluginStateCaf,          
    spGetPluginCurrentPositionCaf,
    spStopPluginCaf,              
    spPausePluginCaf,             
    spRestartPluginCaf,

    spReadPluginCaf,             
    spSeekPluginCaf,
    spGetPluginTotalLengthCaf,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_caf;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(input_caf)

static spBool spOpenInputFileCaf(spCafPluginInstance pinstance, const char *filename)
{
    spChunkSize nread;

    if (pinstance == NULL) {
	spDebug(10, "spOpenInputFileCaf", "instance has'nt been initialized\n");
	return SP_FALSE;
    }
    
    pinstance->fp = NULL;
    pinstance->swap = 0;
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "rb")) == NULL) {
	spDebug(10, "spOpenInputFileCaf", "open failed\n");
	return SP_FALSE;
    }
    spDebug(10, "spOpenInputFileCaf", "filename = %s\n", filename);

    if ((nread = spReadCafHeader(&pinstance->cafHeader, pinstance->fp)) <= 0) {
	spDebug(10, "spOpenInputFileCaf", "spReadCafHeader failed\n");
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
	return SP_FALSE;
    }

    spDebug(50, "spOpenInputFileCaf", "mFormatID: %c%c%c%c\n",
	    pinstance->cafHeader.desc->mFormatID[0], pinstance->cafHeader.desc->mFormatID[1],
	    pinstance->cafHeader.desc->mFormatID[2], pinstance->cafHeader.desc->mFormatID[3]);
    
    spDebug(50, "spOpenInputFileCaf", "mBytesPerPacket = %ld, mChannelsPerFrame = %ld\n",
	    pinstance->cafHeader.desc->mBytesPerPacket, pinstance->cafHeader.desc->mChannelsPerFrame);
    
    pinstance->samp_bit = (8 * pinstance->cafHeader.desc->mBytesPerPacket / pinstance->cafHeader.desc->mChannelsPerFrame);
    if (pinstance->samp_bit == 32
	&& (pinstance->cafHeader.desc->mFormatFlags & SP_CAF_LPCM_FORMAT_FLAG_IS_FLOAT)) {
	pinstance->samp_bit = 33;
    }
    spDebug(50, "spOpenInputFileCaf", "samp_bit = %d\n", pinstance->samp_bit);
    
    if (strneq(pinstance->cafHeader.desc->mFormatID, "ulaw", 4)) {
	spDebug(50, "spOpenInputFileCaf", "format: ulaw\n");
	pinstance->samp_bit = 8;
	pinstance->type_index = SP_CAF_TYPE_ULAW;
    } else if (strneq(pinstance->cafHeader.desc->mFormatID, "alaw", 4)) {
	spDebug(50, "spOpenInputFileCaf", "format: alaw\n");
	pinstance->samp_bit = 8;
	pinstance->type_index = SP_CAF_TYPE_ALAW;
    } else if (strneq(pinstance->cafHeader.desc->mFormatID, "lpcm", 4)
	       && spIsSampleBitSupportedCaf(pinstance->samp_bit) == SP_TRUE) {
	spDebug(50, "spOpenInputFileCaf", "format: lpcm, mFormatFlags = %lx\n", pinstance->cafHeader.desc->mFormatFlags);
	
	if (pinstance->cafHeader.desc->mFormatFlags & SP_CAF_LPCM_FORMAT_FLAG_IS_LITTLE_ENDIAN) {
	    spDebug(50, "spOpenInputFileCaf", "flags: little endian\n");
	    pinstance->swap = (BYTE_ORDER == BIG_ENDIAN); /* swap on big endian environment */
	    pinstance->type_index = SP_CAF_TYPE_LPCM_LE;
	} else {
	    spDebug(50, "spOpenInputFileCaf", "flags: big endian\n");
	    pinstance->swap = (BYTE_ORDER == LITTLE_ENDIAN); /* swap on little endian environment */
	    pinstance->type_index = SP_CAF_TYPE_LPCM_BE;
	}
    } else {
	spDebug(10, "spOpenInputFileCaf", "not Linear PCM CAF\n");
	spCloseFile(pinstance->fp);
	spFreeCafHeader(&pinstance->cafHeader);
	pinstance->fp = NULL;
	return SP_FALSE;
    }

    spDebug(50, "spOpenInputFileCaf", "done\n");
    
    return SP_TRUE;
}

static spBool spCloseInputFileCaf(spCafPluginInstance pinstance)
{
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);

	spFreeCafHeader(&pinstance->cafHeader);
	
	pinstance->fp = NULL;
	pinstance->current_pos = 0;
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spIsSupportedByPluginCaf(const char *filename)
{
    spBool flag;
    spCafPluginInstance pinstance;

    spDebug(10, "spIsSupportedByPluginCaf", "filename = %s\n", filename);
    
    if ((pinstance = (spCafPluginInstance)spInitPluginInstanceCaf("C")) == NULL) {
	return SP_FALSE;
    }
    
    if ((flag = spOpenInputFileCaf(pinstance, filename)) == SP_TRUE) {
	spCloseInputFileCaf(pinstance);
    }
    
    spFreePluginInstanceCaf(pinstance);
    
    spDebug(10, "spIsSupportedByPluginCaf", "flag = %d\n", flag);
    
    return flag;
}

static spPluginError spOpenPluginCaf(void *instance, const char *filename, const char *mode)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    if (spOpenInputFileCaf(pinstance, filename) == SP_FALSE) {
	spDebug(10, "spOpenPluginCaf", "spOpenInputFileCaf failed\n");
	spCloseInputFileCaf(pinstance);
	return SP_PLUGIN_ERROR_OPEN;
    }

    pinstance->sampleRate = pinstance->cafHeader.desc->mSampleRate;
    pinstance->channelsPerFrame = pinstance->cafHeader.desc->mChannelsPerFrame;
    pinstance->bitsPerChannel = pinstance->cafHeader.desc->mBitsPerChannel;

    pinstance->total_duration = spGetCafTotalDuration(&pinstance->cafHeader);

    spParseCafSongInfo(&pinstance->cafHeader, &pinstance->song_info);
    
    spSeekPluginCaf(pinstance, 0);

    spDebug(10, "spOpenPluginCaf", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    return spCloseInputFileCaf(pinstance);
}

static spPluginState spGetPluginStateCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    return (spLong)pinstance->current_pos;
}

static spBool spStopPluginCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginCaf(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginCaf(void *instance)
{

    return SP_TRUE;
}

static long spReadPluginCaf(void *instance, char *data, long length)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    long len;

    if (pinstance->fp == NULL) return -1;

    if (pinstance->current_pos >= pinstance->total_duration) {
	return 0;
    }

    spDebug(80, "spReadPluginCaf", "length = %ld, current_pos = %ld\n",
	    length, (long)pinstance->current_pos);

    if (pinstance->samp_bit == 16) {
	len = freadshort((short *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 64) {
	len = freaddouble((double *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 33) {
	len = freadfloattod((double *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 32) {
	len = freadlong32((long *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 24) {
	len = freadlong24((long *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 8) {
	if (pinstance->type_index == SP_CAF_TYPE_ALAW) {
	    len = freadalaw((short *)data, length, pinstance->fp);
	} else if (pinstance->type_index == SP_CAF_TYPE_ULAW) {
	    len = freadulaw((short *)data, length, pinstance->fp);
	} else {
	    len = freadbyte((short *)data, length, pinstance->fp);
	}
    } else {
	return -1;
    }

    if (len > 0) {
	pinstance->current_pos += (len / pinstance->channelsPerFrame);
    }
    
    spDebug(80, "spReadPluginCaf", "done: len = %ld\n", len);
    
    return len;
}

static spBool spSeekPluginCaf(void *instance, spLong pos)
{
    spFileOffset offset;
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    if (pinstance->fp == NULL) return SP_FALSE;

    offset = spConvertCafTimeToFileOffset(&pinstance->cafHeader, pos, NULL, NULL, NULL);
    spDebug(80, "spSeekPluginCaf", "pos = %ld, offset = %ld\n", (long)pos, (long)offset);
    
    if (spSeekFile(pinstance->fp, offset, SEEK_SET) != 0) {
	spDebug(10, "spSeekPluginCaf", "spSeekFile failed\n");
	return SP_FALSE;
    }
    
    pinstance->current_pos = pos;
    
    return SP_TRUE;
}

static spLong spGetPluginTotalLengthCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    return (spLong)pinstance->total_duration;
}

