#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <pulse/simple.h>
#include <pulse/error.h>
#include <pulse/gccmacro.h>

#include <sp/spBaseLib.h>
#include <sp/spAudio.h>
#include <sp/spRwPluginP.h>

#include <sp/spPluginMain.h>

typedef struct _spPulseSimplePluginInstance
{
    spPluginState state;
    int samp_byte;
    long latency;
    
    int64_t record_total_length;
    int64_t playback_total_length;
    
    pa_buffer_attr paba;
    pa_sample_spec pass;
    
    pa_simple *pasr;
    pa_simple *pasp;

} *spPulseSimplePluginInstance;


/*-- --*/
static spBool spInitPluginPulseSimple(char *lang)
{
    return SP_TRUE;
}

static spBool spFreePluginPulseSimple(void)
{
    return SP_TRUE;
}

static void *spInitPluginInstancePulseSimple(const char *lang)
{
    spPulseSimplePluginInstance p;

    p = xalloc(1, struct _spPulseSimplePluginInstance);
    memset(p, 0, sizeof(struct _spPulseSimplePluginInstance));
    p->samp_byte = 2;
    p->paba.maxlength = (uint32_t)-1;
    p->paba.tlength = (uint32_t)-1;
    p->paba.prebuf = (uint32_t)-1;
    p->paba.minreq = (uint32_t)-1;
    p->paba.fragsize = (uint32_t)-1;
    p->pass.format = PA_SAMPLE_S16LE;
    p->pass.rate = 8000;
    p->pass.channels = 1;

    return p;
}

static spBool spFreePluginInstancePulseSimple(void *instance)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    xfree(p);

    return SP_TRUE;
}

static spBool spGetPluginNumDevicePulseSimple(void *instance, int *num_device)
{
    *num_device = 1;
    return SP_TRUE;
}

static const char *spGetPluginDeviceNamePulseSimple(void *instance, int index)
{
    if (index == 0) {
	return "Default";
    } else {
	return NULL;
    }
}

static spBool spSelectPluginDevicePulseSimple(void *instance, int index)
{
    if (index == 0) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spSetPluginBufferSizePulseSimple(void *instance, int buffer_size)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    p->paba.tlength = (uint32_t)buffer_size;
    p->paba.fragsize = (uint32_t)buffer_size;

    return SP_TRUE;
}
    
static spBool spGetPluginBufferSizePulseSimple(void *instance, int *buffer_size)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    *buffer_size = (int)p->paba.tlength;

    return SP_TRUE;
}

static spBool spSetPluginSampleBitPulseSimple(void *instance, int samp_bit)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    p->samp_byte = samp_bit / 8;

    return SP_TRUE;
}

static spBool spGetPluginSampleBitPulseSimple(void *instance, int *samp_bit)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    *samp_bit = 8 * p->samp_byte;

    return SP_TRUE;
}
    
static spBool spSetPluginChannelPulseSimple(void *instance, int num_channel)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    p->pass.channels = MAX(num_channel, 1);

    return SP_TRUE;
}

static spBool spGetPluginChannelPulseSimple(void *instance, int *num_channel)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    *num_channel = p->pass.channels;

    return SP_TRUE;
}
    
static spBool spSetPluginSampleRatePulseSimple(void *instance, double samp_rate)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    p->pass.rate = (uint32_t)spRound(samp_rate);

    return SP_TRUE;
}

static spBool spGetPluginSampleRatePulseSimple(void *instance, double *samp_rate)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    *samp_rate = (double)p->pass.rate;

    return SP_TRUE;
}

static spPluginError spOpenPluginPulseSimple(void *instance, const char *filename, const char *mode)
{
    int error;
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    if (mode[0] == 'r') {
	if ((p->pasr = pa_simple_new(NULL, "spPlugin", PA_STREAM_RECORD, NULL, "record",
				     &p->pass, NULL, &p->paba, &error)) == NULL) {
	    spDebug(10, "spOpenPluginPulseSimple", "pa_simple_new() for PA_STREAM_RECORD failed: %s\n", pa_strerror(error));
	    p->state = SP_PLUGIN_STATE_ERROR;
	    return SP_PLUGIN_ERROR_OPEN;
	}
	p->record_total_length = 0;
    }
    
    if (mode[0] == 'w' || mode[1] == 'w') {
	pa_usec_t latency;
	 
	if ((p->pasp = pa_simple_new(NULL, "spPlugin", PA_STREAM_PLAYBACK, NULL, "playback",
				     &p->pass, NULL, &p->paba, &error)) == NULL) {
	    spDebug(10, "spOpenPluginPulseSimple", "pa_simple_new() for PA_STREAM_PLAYBACK failed: %s\n", pa_strerror(error));
	    p->state = SP_PLUGIN_STATE_ERROR;
	    return SP_PLUGIN_ERROR_OPEN;
	}
        if ((latency = pa_simple_get_latency(p->pasp, &error)) != (pa_usec_t)-1) {
	    p->latency = (long)pa_usec_to_bytes(latency, &p->pass) / (long)p->samp_byte;
            spDebug(10, "spOpenPluginPulseSimple", "pa_simple_get_latency() result = %ld\n", p->latency);
	} else {
            spDebug(10, "spOpenPluginPulseSimple", "pa_simple_get_latency() failed: %s\n", pa_strerror(error));
        }	    
	p->playback_total_length = 0;
    }
    p->state = SP_PLUGIN_STATE_STOP;
    
    spDebug(10, "spOpenPluginPulseSimple", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginPulseSimple(void *instance)
{
    int error;
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    if (p->pasp != NULL) {
	if (p->playback_total_length > 0) {
	    /* Make sure that every single sample was played */
	    if (pa_simple_drain(p->pasp, &error) < 0) {
		spDebug(10, "spClosePluginPulseSimple", "pa_simple_drain() failed: %s\n", pa_strerror(error));
	    }
	}
    	pa_simple_free(p->pasp);
	p->pasp = NULL;
	p->playback_total_length = 0;
    }
    if (p->pasr != NULL) {
	pa_simple_free(p->pasr);
	p->pasr = NULL;
	p->record_total_length = 0;
    }
    p->state = SP_PLUGIN_STATE_STOP;

    spDebug(10, "spClosePluginPulseSimple", "done\n");
    
    return SP_TRUE;
}

static spPluginState spGetPluginStatePulseSimple(void *instance)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    return p->state;
}

static spLong spGetPluginCurrentPositionPulseSimple(void *instance)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    return (p->playback_total_length / (int64_t)p->pass.channels) - (int64_t)p->latency;
}

static spBool spStopPluginPulseSimple(void *instance)
{
    int error;
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    if (p->playback_total_length > 0) {
	if (pa_simple_flush(p->pasp, &error) < 0) {
	    spDebug(10, "spStopPluginPulseSimple", "pa_simple_flush() failed: %s\n", pa_strerror(error));
	    p->state = SP_PLUGIN_STATE_ERROR;
	    return SP_FALSE;
	}
    }
    p->playback_total_length = 0;
    p->state = SP_PLUGIN_STATE_STOP;

    return SP_TRUE;
}

static spBool spPausePluginPulseSimple(void *instance)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;
    
    p->state = SP_PLUGIN_STATE_PAUSE;
    return SP_TRUE;
}

static spBool spRestartPluginPulseSimple(void *instance)
{
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;
    
    p->state = SP_PLUGIN_STATE_START;
    return SP_TRUE;
}

static long spWritePluginPulseSimple(void *instance, char *data, long length)
{
    int error;
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    if (p->pasp == NULL) return -1;

    spDebug(10, "spWritePluginPulseSimple", "length = %ld\n", length);
    
    if (pa_simple_write(p->pasp, data, (size_t)(length * p->samp_byte), &error) < 0) {
	spDebug(10, "spWritePluginPulseSimple", "pa_simple_write() failed: %s\n", pa_strerror(error));
	p->state = SP_PLUGIN_STATE_ERROR;
	return -1;
    }
    p->playback_total_length += (int64_t)length;
    p->state = SP_PLUGIN_STATE_START;
    
    spDebug(10, "spWritePluginPulseSimple", "done\n");
    
    return length;
}

static spBool spFlushPluginPulseSimple(void *instance)
{
    int error;
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;

    if (p->pasp != NULL) {
	if (p->playback_total_length > 0) {
	    /* Make sure that every single sample was played */
	    if (pa_simple_drain(p->pasp, &error) < 0) {
		spDebug(10, "spFlushPluginPulseSimple", "pa_simple_drain() failed: %s\n", pa_strerror(error));
	    }
	}
    	pa_simple_free(p->pasp);
	p->pasp = NULL;

	p->state = SP_PLUGIN_STATE_STOP;
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static long spReadPluginPulseSimple(void *instance, char *data, long length)
{
    int error;
    spPulseSimplePluginInstance p = (spPulseSimplePluginInstance )instance;
    
    if (p->pasr == NULL) return -1;

    spDebug(10, "spReadPluginPulseSimple", "length = %ld\n", length);
    
    if (pa_simple_read(p->pasr, data, (size_t)(length * p->samp_byte), &error) < 0) {
	spDebug(10, "spReadPluginPulseSimple", "pa_simple_read() failed: %s\n", pa_strerror(error));
	p->state = SP_PLUGIN_STATE_ERROR;
	return -1;
    }
    p->record_total_length += (int64_t)length;
    p->state = SP_PLUGIN_STATE_START;

    spDebug(10, "spReadPluginPulseSimple", "done\n");
    
    return length;
}

static spRwPluginRec sp_rw_plugin_pulse_simple = {
    NULL,
    NULL,

    SP_PLUGIN_RW,
    "PulseSimple",
    2,
    SP_PLUGIN_PRIORITY_NULL,
    0L,
    spInitPluginPulseSimple,
    spFreePluginPulseSimple,
    "PulseAudio (Simple) Plugin",
    "PulseAudio (Simple) Plugin Version 0.2\nCopyright(C) Hideki Banno",

    spInitPluginInstancePulseSimple,
    spFreePluginInstancePulseSimple,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_AUDIO,
    NULL,
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    0,
    NULL,
    NULL,
    NULL,

    spGetPluginNumDevicePulseSimple,
    spGetPluginDeviceNamePulseSimple,
    spSelectPluginDevicePulseSimple,
    spSetPluginBufferSizePulseSimple,  
    spGetPluginBufferSizePulseSimple,
    NULL,      
    NULL,      
    
    spSetPluginSampleBitPulseSimple,   
    spGetPluginSampleBitPulseSimple,   
    spSetPluginChannelPulseSimple,     
    spGetPluginChannelPulseSimple,     
    spSetPluginSampleRatePulseSimple,  
    spGetPluginSampleRatePulseSimple,  
    NULL,   
    NULL,
    
    spOpenPluginPulseSimple,              
    spClosePluginPulseSimple,             
    spGetPluginStatePulseSimple,          
    spGetPluginCurrentPositionPulseSimple,
    spStopPluginPulseSimple,              
    spPausePluginPulseSimple,             
    spRestartPluginPulseSimple,

    spWritePluginPulseSimple,             
    spFlushPluginPulseSimple,
    NULL,

    spReadPluginPulseSimple,             
    NULL,
    NULL,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_rw_plugin_pulse_simple;
}
